// CycloneDriver.cpp: implementation of the CCycloneDriver class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "CycloneDriver.h"
#include "CyApi.h"
#include <WinIoCtl.h>
#include "../CycloneDI.h"


using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
const char* CCycloneDriver::g_LinkName="\\\\.\\CycloneDevice0";

CCycloneDriver::CCycloneDriver()  
{
	m_hDevice=INVALID_HANDLE_VALUE;
	m_bNotifyThreadStartup=false;
}

CCycloneDriver::~CCycloneDriver()
{
	
	if (INVALID_HANDLE_VALUE!=m_NotifyAvail)
	{
		CloseHandle(m_NotifyAvail);
	}

	

}

CCycloneDriver& CCycloneDriver::Instance()
{
	static CCycloneDriver obj;
	return obj;
}

int CCycloneDriver::OpenDriver(CNotification* p)
{
	if (IsOpen())
	{
		return CYAPI_ERR_ALREADY_OPENNED;
	}

	//
	// Open the cyclone driver
	//
	m_hDevice=CreateFile(g_LinkName,
					  GENERIC_READ | GENERIC_WRITE,
					  FILE_SHARE_READ,
					  NULL,
					  OPEN_EXISTING,
					  0,
					  NULL);
	if (!IsOpen())
	{
		return CYAPI_INTERNAL_ERROR;
	}

	if (!AcquireShm())
	{
		CloseDriver();
		return CYAPI_INTERNAL_ERROR;
	}

	m_pNotify=p;

	//
	// Create the semaphore to synchronize the driver events
	//
	m_NotifyAvail=OpenEvent(SYNCHRONIZE, FALSE,DRIVER_EVENT_NAME);
	if (INVALID_HANDLE_VALUE==m_NotifyAvail)
	{
		CloseDriver();
		return CYAPI_INTERNAL_ERROR;
	}

	EnableNotify();
	
	//
	// Start the Event notification thread
	//
	m_bNotifyThreadStartup=false;
	CMclThread startThread(this);
	return CYAPI_SUCCESS;
}

bool CCycloneDriver::AcquireShm()
{
	if (!IsOpen())
	{
		return false;
	}
	//
	// Get the internal shared memory address
	//
	DWORD nRet;
	MemoryResult Memory;
	BOOL status = 
		DeviceIoControl(m_hDevice, 
						IOCTL_CYCLONE_GETMEMORY, 
						NULL, 
						0,
						&Memory,
						sizeof(Memory),
						&nRet, 
						NULL);

	if ( !status )
	{
		return false;
	}

	m_pShm=Memory.m_ret;

	if (IsBadReadPtr(m_pShm,sizeof(SHARED_BUFFER))||
		IsBadWritePtr(m_pShm,sizeof(SHARED_BUFFER)))
	{
		return false;
	}

	PSHARED_BUFFER pShm=(PSHARED_BUFFER)m_pShm;
	if (pShm->ulSig!=0xbeefdead)
	{
		return false;
	}
	return true;
}

void CCycloneDriver::CloseDriver()
{
	DisableNotify();
	if (m_bNotifyThreadStartup)
	{
		StopNotifyThread();
	}

	if (IsOpen())
	{
		CloseHandle(m_hDevice);
		m_hDevice=INVALID_HANDLE_VALUE;
	}
}


bool CCycloneDriver::ReadOneRingBufferEntry(NOTIFYBUFFER& ring_buffer, std::vector<char>& entry)
{
	if(InterlockedCompareExchange(&ring_buffer.count, 0, 0) == 0) return false;

	// Buffer format is DWORD length, followed by data
	DWORD entry_size = *(DWORD*)(ring_buffer.buffer + ring_buffer.read);
//	ASSERT(entry_size < ring_buffer.size);

	// If entry not yet ready, the size won't have been set
	if(!entry_size) return false;
	
	// Safe now to mark it as used - one way or another we're going to clear this baby out
	*(DWORD*)(ring_buffer.buffer + ring_buffer.read) = 0;

	// Move on to actual data and pull it to a vector
	ring_buffer.read = (ring_buffer.read + sizeof(entry_size)) % ring_buffer.size;

	entry.resize(entry_size);
	char* start = ring_buffer.buffer + ring_buffer.read;
	char* end = ring_buffer.buffer + ring_buffer.read + entry_size;

	if(end > ring_buffer.buffer + ring_buffer.size)
	{
		int remaining = ring_buffer.size - ring_buffer.read;

		std::copy(start, ring_buffer.buffer + ring_buffer.size, entry.begin());
		std::fill(start, ring_buffer.buffer + ring_buffer.size, 0);
		std::copy(ring_buffer.buffer, ring_buffer.buffer + entry_size - remaining, entry.begin() + remaining);
		std::fill(ring_buffer.buffer, ring_buffer.buffer + entry_size - remaining, 0);
	}
	else
	{
		std::copy(start, end, entry.begin());
		std::fill(start, end, 0);
	}

	// Allow this space to be reused.
	ring_buffer.read = (ring_buffer.read + entry_size) % ring_buffer.size;
	InterlockedExchangeAdd(&ring_buffer.count, -entry_size - sizeof(DWORD));
	return true;
}

unsigned CCycloneDriver::ThreadHandlerProc(void)
{
	vector<char> processBuffer;
	PSHARED_BUFFER pShm= (PSHARED_BUFFER)m_pShm;
	m_bNotifyThreadStartup=true;
	while (true)
	{
		HANDLE handles[2];

		// the current object...
		handles[0] = m_EventStop;

		// the parameter object...
		handles[1] = m_NotifyAvail;

		// wait for the objects...
		DWORD waitStatus =::WaitForMultipleObjects( 2, handles, FALSE, INFINITE);

 		if (!CMclWaitSucceeded(waitStatus,2))
		{
			//
			// Program internal error. Stop the thread
			//
			break;
		}
			
		if (CMclWaitSucceededIndex(waitStatus)==WAIT_OBJECT_0)
		{
			//
			// Receive the stop event, quit the thread
			//
			break;
		}

		//
		// Process the notify event;
		//
		while (ReadOneRingBufferEntry(pShm->notify_buffer,processBuffer))
		{
			m_pNotify->Notify(processBuffer.begin(),processBuffer.size());
		}
	}
	m_EventStopAck.Set();
	m_bNotifyThreadStartup=false;
	return 0;
}

bool CCycloneDriver::EnableNotify()
{
	if (!IsOpen())
	{
		return false;
	}
	//
	// Get the internal shared memory address
	//
	DWORD nRet;
	BOOL status = 
		DeviceIoControl(m_hDevice, 
						IOCTL_CYCLONE_ENABLE_NOTIFY, 
						NULL, 
						0,
						NULL,
						0,
						&nRet, 
						NULL);

	return status==TRUE;
}

bool CCycloneDriver::DisableNotify()
{
	if (!IsOpen())
	{
		return false;
	}
	//
	// Get the internal shared memory address
	//
	DWORD nRet;
	BOOL status = 
		DeviceIoControl(m_hDevice, 
						IOCTL_CYCLONE_DISABLE_NOTIFY, 
						NULL, 
						0,
						NULL,
						0,
						&nRet, 
						NULL);

	return status==TRUE;
}
